import unittest

from pgmpy.independencies import Independencies, IndependenceAssertion


class TestIndependenceAssertion(unittest.TestCase):
    def setUp(self):
        self.assertion = IndependenceAssertion()

    def test_return_list_if_str(self):
        self.assertListEqual(self.assertion._return_list_if_str("U"), ["U"])
        self.assertListEqual(self.assertion._return_list_if_str(["U", "V"]), ["U", "V"])

    def test_get_assertion(self):
        self.assertTupleEqual(
            IndependenceAssertion("U", "V", "Z").get_assertion(), ({"U"}, {"V"}, {"Z"})
        )
        self.assertTupleEqual(
            IndependenceAssertion("U", "V").get_assertion(), ({"U"}, {"V"}, set())
        )

    def test_init(self):
        self.assertion1 = IndependenceAssertion("U", "V", "Z")
        self.assertSetEqual(self.assertion1.event1, {"U"})
        self.assertSetEqual(self.assertion1.event2, {"V"})
        self.assertSetEqual(self.assertion1.event3, {"Z"})
        self.assertion1 = IndependenceAssertion(["U", "V"], ["Y", "Z"], ["A", "B"])
        self.assertSetEqual(self.assertion1.event1, {"U", "V"})
        self.assertSetEqual(self.assertion1.event2, {"Y", "Z"})
        self.assertSetEqual(self.assertion1.event3, {"A", "B"})

    def test_init_exceptions(self):
        self.assertRaises(ValueError, IndependenceAssertion, event2=["U"], event3="V")
        self.assertRaises(ValueError, IndependenceAssertion, event2=["U"])
        self.assertRaises(ValueError, IndependenceAssertion, event3=["Z"])
        self.assertRaises(ValueError, IndependenceAssertion, event1=["U"])
        self.assertRaises(ValueError, IndependenceAssertion, event1=["U"], event3=["Z"])

    def tearDown(self):
        del self.assertion


class TestIndependeciesAssertionEq(unittest.TestCase):
    def setUp(self):
        self.i1 = IndependenceAssertion("a", "b", "c")
        self.i2 = IndependenceAssertion("a", "b")
        self.i3 = IndependenceAssertion("a", ["b", "c", "d"])
        self.i4 = IndependenceAssertion("a", ["b", "c", "d"], "e")
        self.i5 = IndependenceAssertion("a", ["d", "c", "b"], "e")
        self.i6 = IndependenceAssertion("a", ["d", "c"], ["e", "b"])
        self.i7 = IndependenceAssertion("a", ["c", "d"], ["b", "e"])
        self.i8 = IndependenceAssertion("a", ["f", "d"], ["b", "e"])
        self.i9 = IndependenceAssertion("a", ["d", "k", "b"], "e")
        self.i10 = IndependenceAssertion(["k", "b", "d"], "a", "e")

    def test_eq1(self):
        self.assertFalse(self.i1 == "a")
        self.assertFalse(self.i2 == 1)
        self.assertFalse(self.i4 == [2, "a"])
        self.assertFalse(self.i6 == "c")

    def test_eq2(self):
        self.assertFalse(self.i1 == self.i2)
        self.assertFalse(self.i1 == self.i3)
        self.assertFalse(self.i2 == self.i4)
        self.assertFalse(self.i3 == self.i6)

    def test_eq3(self):
        self.assertTrue(self.i4 == self.i5)
        self.assertTrue(self.i6 == self.i7)
        self.assertFalse(self.i7 == self.i8)
        self.assertFalse(self.i4 == self.i9)
        self.assertFalse(self.i5 == self.i9)
        self.assertTrue(self.i10 == self.i9)
        self.assertTrue(self.i10 != self.i8)

    def tearDown(self):
        del self.i1
        del self.i2
        del self.i3
        del self.i4
        del self.i5
        del self.i6
        del self.i7
        del self.i8
        del self.i9
        del self.i10


class TestIndependencies(unittest.TestCase):
    def setUp(self):
        self.Independencies = Independencies()
        self.Independencies3 = Independencies(
            ["a", ["b", "c", "d"], ["e", "f", "g"]], ["c", ["d", "e", "f"], ["g", "h"]]
        )
        self.Independencies4 = Independencies(
            [["f", "d", "e"], "c", ["h", "g"]], [["b", "c", "d"], "a", ["f", "g", "e"]]
        )
        self.Independencies5 = Independencies(
            ["a", ["b", "c", "d"], ["e", "f", "g"]], ["c", ["d", "e", "f"], "g"]
        )

    def test_init(self):
        self.Independencies1 = Independencies(["X", "Y", "Z"])
        self.assertEqual(self.Independencies1, Independencies(["X", "Y", "Z"]))
        self.Independencies2 = Independencies()
        self.assertEqual(self.Independencies2, Independencies())

    def test_add_assertions(self):
        self.Independencies1 = Independencies(["X", "Y", "Z"])
        self.assertEqual(self.Independencies1, Independencies(["X", "Y", "Z"]))
        self.Independencies2 = Independencies(["A", "B", "C"], ["D", "E", "F"])
        self.assertEqual(
            self.Independencies2, Independencies(["A", "B", "C"], ["D", "E", "F"])
        )

    def test_get_assertions(self):
        self.Independencies1 = Independencies(["X", "Y", "Z"])
        self.assertEqual(
            self.Independencies1.independencies, self.Independencies1.get_assertions()
        )
        self.Independencies2 = Independencies(["A", "B", "C"], ["D", "E", "F"])
        self.assertEqual(
            self.Independencies2.independencies, self.Independencies2.get_assertions()
        )

    def test_closure(self):
        ind1 = Independencies(("A", ["B", "C"], "D"))
        self.assertEqual(
            ind1.closure(),
            Independencies(
                ("A", ["B", "C"], "D"),
                ("A", "B", ["C", "D"]),
                ("A", "C", ["B", "D"]),
                ("A", "B", "D"),
                ("A", "C", "D"),
            ),
        )
        ind2 = Independencies(("W", ["X", "Y", "Z"]))
        self.assertEqual(
            ind2.closure(),
            Independencies(
                ("W", "Y"),
                ("W", "Y", "X"),
                ("W", "Y", "Z"),
                ("W", "Y", ["X", "Z"]),
                ("W", ["Y", "X"]),
                ("W", "X", ["Y", "Z"]),
                ("W", ["X", "Z"], "Y"),
                ("W", "X"),
                ("W", ["X", "Z"]),
                ("W", ["Y", "Z"], "X"),
                ("W", ["Y", "X", "Z"]),
                ("W", "X", "Z"),
                ("W", ["Y", "Z"]),
                ("W", "Z", "X"),
                ("W", "Z"),
                ("W", ["Y", "X"], "Z"),
                ("W", "X", "Y"),
                ("W", "Z", ["Y", "X"]),
                ("W", "Z", "Y"),
            ),
        )
        ind3 = Independencies(
            ("c", "a", ["b", "e", "d"]),
            (["e", "c"], "b", ["a", "d"]),
            (["b", "d"], "e", "a"),
            ("e", ["b", "d"], "c"),
            ("e", ["b", "c"], "d"),
            (["e", "c"], "a", "b"),
        )
        self.assertEqual(len(ind3.closure().get_assertions()), 78)

    def test_entails(self):
        ind1 = Independencies([["A", "B"], ["C", "D"], "E"])
        ind2 = Independencies(["A", "C", "E"])
        self.assertTrue(ind1.entails(ind2))
        self.assertFalse(ind2.entails(ind1))
        ind3 = Independencies(("W", ["X", "Y", "Z"]))
        self.assertTrue(ind3.entails(ind3.closure()))
        self.assertTrue(ind3.closure().entails(ind3))

    def test_is_equivalent(self):
        ind1 = Independencies(["X", ["Y", "W"], "Z"])
        ind2 = Independencies(["X", "Y", "Z"], ["X", "W", "Z"])
        ind3 = Independencies(["X", "Y", "Z"], ["X", "W", "Z"], ["X", "Y", ["W", "Z"]])
        self.assertFalse(ind1.is_equivalent(ind2))
        self.assertTrue(ind1.is_equivalent(ind3))

    def test_eq(self):
        self.assertTrue(self.Independencies3 == self.Independencies4)
        self.assertFalse(self.Independencies3 != self.Independencies4)
        self.assertTrue(self.Independencies3 != self.Independencies5)
        self.assertFalse(self.Independencies4 == self.Independencies5)
        self.assertFalse(Independencies() == Independencies(["A", "B", "C"]))
        self.assertFalse(Independencies(["A", "B", "C"]) == Independencies())
        self.assertTrue(Independencies() == Independencies())

    def tearDown(self):
        del self.Independencies
        del self.Independencies3
        del self.Independencies4
        del self.Independencies5


if __name__ == "__main__":
    unittest.main()



from __future__ import division

import types

import numpy as np
import scipy.integrate as integrate

from pgmpy.factors.base import BaseFactor
from pgmpy.factors.distributions import GaussianDistribution, CustomDistribution


class ContinuousFactor(BaseFactor):
    """
    Base class for factors representing various multivariate
    representations.
    """

    def __init__(self, variables, pdf, *args, **kwargs):
        """
        Parameters
        ----------
        variables: list or array-like
            The variables for wich the distribution is defined.

        pdf: function
            The probability density function of the distribution.

        Examples
        --------
        >>> import numpy as np
        >>> from scipy.special import beta
        >>> from pgmpy.factors.continuous import ContinuousFactor
        # Two variable dirichlet distribution with alpha = (1,2)
        >>> def dirichlet_pdf(x, y):
        ...     return (np.power(x, 1) * np.power(y, 2)) / beta(x, y)
        >>> dirichlet_factor = ContinuousFactor(['x', 'y'], dirichlet_pdf)
        >>> dirichlet_factor.scope()
        ['x', 'y']
        >>> dirichlet_factor.assignment(5,6)
        226800.0
        """
        if not isinstance(variables, (list, tuple, np.ndarray)):
            raise TypeError(
                "variables: Expected type list or array-like, "
                "got type {var_type}".format(var_type=type(variables))
            )

        if len(set(variables)) != len(variables):
            raise ValueError("Variable names cannot be same.")

        variables = list(variables)

        if isinstance(pdf, str):
            if pdf == "gaussian":
                self.distribution = GaussianDistribution(
                    variables=variables,
                    mean=kwargs["mean"],
                    covariance=kwargs["covariance"],
                )
            else:
                raise NotImplementedError(
                    "{dist} distribution not supported.",
                    "Please use CustomDistribution".format(dist=pdf),
                )

        elif isinstance(pdf, CustomDistribution):
            self.distribution = pdf

        elif callable(pdf):
            self.distribution = CustomDistribution(
                variables=variables, distribution=pdf
            )

        else:
            raise ValueError(
                "pdf: Expected type: str or function, ",
                "Got: {instance}".format(instance=type(variables)),
            )

    @property
    def pdf(self):
        """
        Returns the pdf of the ContinuousFactor.
        """
        return self.distribution.pdf

    @property
    def variable(self):
        return self.scope()[0]

    def scope(self):
        """
        Returns the scope of the factor.

        Returns
        -------
        list: List of variable names in the scope of the factor.

        Examples
        --------
        >>> from pgmpy.factors.continuous import ContinuousFactor
        >>> from scipy.stats import multivariate_normal
        >>> normal_pdf = lambda x: multivariate_normal(x, [0, 0], [[1, 0], [0, 1]])
        >>> phi = ContinuousFactor(['x1', 'x2'], normal_pdf)
        >>> phi.scope()
        ['x1', 'x2']
        """
        return self.distribution.variables

    def get_evidence(self):
        return self.scope()[1:]

    def assignment(self, *args):
        """
        Returns a list of pdf assignments for the corresponding values.

        Parameters
        ----------
        *args: values
            Values whose assignment is to be computed.

        Examples
        --------
        >>> from pgmpy.factors.continuous import ContinuousFactor
        >>> from scipy.stats import multivariate_normal
        >>> normal_pdf = lambda x1, x2: multivariate_normal.pdf((x1, x2), [0, 0], [[1, 0], [0, 1]])
        >>> phi = ContinuousFactor(['x1', 'x2'], normal_pdf)
        >>> phi.assignment(1, 2)
        0.013064233284684921
        """
        return self.distribution.assignment(*args)

    def copy(self):
        """
        Return a copy of the distribution.

        Returns
        -------
        ContinuousFactor object: copy of the distribution

        Examples
        --------
        >>> import numpy as np
        >>> from scipy.special import beta
        >>> from pgmpy.factors.continuous import ContinuousFactor
        # Two variable dirichlet distribution with alpha = (1,2)
        >>> def dirichlet_pdf(x, y):
        ...     return (np.power(x, 1) * np.power(y, 2)) / beta(x, y)
        >>> dirichlet_factor = ContinuousFactor(['x', 'y'], dirichlet_pdf)
        >>> dirichlet_factor.variables
        ['x', 'y']
        >>> copy_factor = dirichlet_factor.copy()
        >>> copy_factor.variables
        ['x', 'y']
        """
        return ContinuousFactor(self.scope(), self.distribution.copy())

    def discretize(self, method, *args, **kwargs):
        """
        Discretizes the continuous distribution into discrete
        probability masses using various methods.

        Parameters
        ----------
        method : A Discretizer Class from pgmpy.discretize

        *args, **kwargs:
            The parameters to be given to the Discretizer Class.

        Returns
        -------
        An n-D array or a DiscreteFactor object according to the discretiztion
        method used.

        Examples
        --------
        >>> import numpy as np
        >>> from scipy.special import beta
        >>> from pgmpy.factors.continuous import ContinuousFactor
        >>> from pgmpy.factors.continuous import RoundingDiscretizer
        >>> def dirichlet_pdf(x, y):
        ...     return (np.power(x, 1) * np.power(y, 2)) / beta(x, y)
        >>> dirichlet_factor = ContinuousFactor(['x', 'y'], dirichlet_pdf)
        >>> dirichlet_factor.discretize(RoundingDiscretizer, low=1, high=2, cardinality=5)
        # TODO: finish this
        """
        return method(self, *args, **kwargs).get_discrete_values()

    def reduce(self, values, inplace=True):
        """
        Reduces the factor to the context of the given variable values.

        Parameters
        ----------
        values: list, array-like
            A list of tuples of the form (variable_name, variable_value).

        inplace: boolean
            If inplace=True it will modify the factor itself, else would return
            a new ContinuosFactor object.

        Returns
        -------
        ContinuousFactor or None: if inplace=True (default) returns None
                                  if inplace=False returns a new ContinuousFactor instance.

        Examples
        --------
        >>> import numpy as np
        >>> from scipy.special import beta
        >>> from pgmpy.factors.continuous import ContinuousFactor
        >>> def custom_pdf(x, y, z):
        ...     return z*(np.power(x, 1) * np.power(y, 2)) / beta(x, y)
        >>> custom_factor = ContinuousFactor(['x', 'y', 'z'], custom_pdf)
        >>> custom_factor.variables
        ['x', 'y', 'z']
        >>> custom_factor.assignment(1, 2, 3)
        24.0

        >>> custom_factor.reduce([('y', 2)])
        >>> custom_factor.variables
        ['x', 'z']
        >>> custom_factor.assignment(1, 3)
        24.0
        """
        phi = self if inplace else self.copy()

        phi.distribution = phi.distribution.reduce(values, inplace=False)
        if not inplace:
            return phi

    def marginalize(self, variables, inplace=True):
        """
        Marginalize the factor with respect to the given variables.

        Parameters
        ----------
        variables: list, array-like
            List of variables with respect to which factor is to be maximized.

        inplace: boolean
            If inplace=True it will modify the factor itself, else would return
            a new ContinuousFactor instance.

        Returns
        -------
        DiscreteFactor or None: if inplace=True (default) returns None
                        if inplace=False returns a new ContinuousFactor instance.

        Examples
        --------
        >>> from pgmpy.factors.continuous import ContinuousFactor
        >>> from scipy.stats import multivariate_normal
        >>> std_normal_pdf = lambda *x: multivariate_normal.pdf(x, [0, 0], [[1, 0], [0, 1]])
        >>> std_normal = ContinuousFactor(['x1', 'x2'], std_normal_pdf)
        >>> std_normal.scope()
        ['x1', 'x2']
        >>> std_normal.assignment([1, 1])
        0.058549831524319168
        >>> std_normal.marginalize(['x2'])
        >>> std_normal.scope()
        ['x1']
        >>> std_normal.assignment(1)

        """
        phi = self if inplace else self.copy()
        phi.distribution = phi.distribution.marginalize(variables, inplace=False)

        if not inplace:
            return phi

    def normalize(self, inplace=True):
        """
        Normalizes the pdf of the continuous factor so that it integrates to
        1 over all the variables.

        Parameters
        ----------
        inplace: boolean
            If inplace=True it will modify the factor itself, else would return
            a new factor.

        Returns
        -------
        ContinuousFactor or None:
             if inplace=True (default) returns None
             if inplace=False returns a new ContinuousFactor instance.

        Examples
        --------
        >>> from pgmpy.factors.continuous import ContinuousFactor
        >>> from scipy.stats import multivariate_normal
        >>> std_normal_pdf = lambda x: 2 * multivariate_normal.pdf(x, [0, 0], [[1, 0], [0, 1]])
        >>> std_normal = ContinuousFactor(['x1', 'x2'], std_normal_pdf)
        >>> std_normal.assignment(1, 1)
        0.117099663049
        >>> std_normal.normalize()
        >>> std_normal.assignment(1, 1)
        0.0585498315243

        """
        phi = self if inplace else self.copy()
        phi.distribution.normalize(inplace=True)

        if not inplace:
            return phi

    def is_valid_cpd(self):
        return self.distribution.is_valid_cpd()

    def _operate(self, other, operation, inplace=True):
        """
        Gives the ContinuousFactor operation (product or divide) with
        the other factor.

        Parameters
        ----------
        other: ContinuousFactor
            The ContinuousFactor to be multiplied.

        operation: String
            'product' for multiplication operation and 'divide' for
            division operation.

        inplace: boolean
            If inplace=True it will modify the factor itself, else would return
            a new factor.

        Returns
        -------
        ContinuousFactor or None:
                        if inplace=True (default) returns None
                        if inplace=False returns a new `DiscreteFactor` instance.

        """
        if not isinstance(other, ContinuousFactor):
            raise TypeError(
                "ContinuousFactor objects can only be multiplied ",
                "or divided with another ContinuousFactor object. ",
                "Got {other_type}, expected: ContinuousFactor.".format(
                    other_type=type(other)
                ),
            )

        phi = self if inplace else self.copy()
        phi.distribution = phi.distribution._operate(
            other=other.distribution, operation=operation, inplace=False
        )

        if not inplace:
            return phi

    def product(self, other, inplace=True):
        """
        Gives the ContinuousFactor product with the other factor.

        Parameters
        ----------
        other: ContinuousFactor
            The ContinuousFactor to be multiplied.

        Returns
        -------
        ContinuousFactor or None:
                        if inplace=True (default) returns None
                        if inplace=False returns a new `ContinuousFactor` instance.

        Examples
        --------
        >>> from pgmpy.factors.continuous import ContinuousFactor
        >>> from scipy.stats import multivariate_normal
        >>> sn_pdf1 = lambda x: multivariate_normal.pdf([x], [0], [[1]])
        >>> sn_pdf2 = lambda x1,x2: multivariate_normal.pdf([x1, x2], [0, 0], [[1, 0], [0, 1]])
        >>> sn1 = ContinuousFactor(['x2'], sn_pdf1)
        >>> sn2 = ContinuousFactor(['x1', 'x2'], sn_pdf2)

        >>> sn3 = sn1.product(sn2, inplace=False)
        >>> sn3.assignment(0, 0)
        0.063493635934240983

        >>> sn3 = sn1 * sn2
        >>> sn3.assignment(0, 0)
        0.063493635934240983
        """
        return self._operate(other, "product", inplace)

    def divide(self, other, inplace=True):
        """
        Gives the ContinuousFactor divide with the other factor.

        Parameters
        ----------
        other: ContinuousFactor
            The ContinuousFactor to be divided.

        Returns
        -------
        ContinuousFactor or None:
                        if inplace=True (default) returns None
                        if inplace=False returns a new `ContinuousFactor` instance.

        Examples
        --------
        >>> from pgmpy.factors.continuous import ContinuousFactor
        >>> from scipy.stats import multivariate_normal
        >>> sn_pdf1 = lambda x: multivariate_normal.pdf([x], [0], [[1]])
        >>> sn_pdf2 = lambda x1,x2: multivariate_normal.pdf([x1, x2], [0, 0], [[1, 0], [0, 1]])
        >>> sn1 = ContinuousFactor(['x2'], sn_pdf1)
        >>> sn2 = ContinuousFactor(['x1', 'x2'], sn_pdf2)

        >>> sn4 = sn2.divide(sn1, inplace=False)
        >>> sn4.assignment(0, 0)
        0.3989422804014327

        >>> sn4 = sn2 / sn1
        >>> sn4.assignment(0, 0)
        0.3989422804014327
        """
        if set(other.scope()) - set(self.scope()):
            raise ValueError("Scope of divisor should be a subset of dividend")

        return self._operate(other, "divide", inplace)

    def __mul__(self, other):
        return self.product(other, inplace=False)

    def __rmul__(self, other):
        return self.__mul__(other)

    def __truediv__(self, other):
        return self.divide(other, inplace=False)

    __div__ = __truediv__



# -*- coding: utf-8 -*-

from itertools import chain

import numpy as np

from pgmpy.estimators import ParameterEstimator
from pgmpy.factors.discrete import TabularCPD
from pgmpy.models import BayesianModel


class BayesianEstimator(ParameterEstimator):
    def __init__(self, model, data, **kwargs):
        """
        Class used to compute parameters for a model using Bayesian Parameter Estimation.
        See `MaximumLikelihoodEstimator` for constructor parameters.
        """
        if not isinstance(model, BayesianModel):
            raise NotImplementedError(
                "Bayesian Parameter Estimation is only implemented for BayesianModel"
            )

        super(BayesianEstimator, self).__init__(model, data, **kwargs)

    def get_parameters(
        self, prior_type="BDeu", equivalent_sample_size=5, pseudo_counts=None
    ):
        """
        Method to estimate the model parameters (CPDs).

        Parameters
        ----------
        prior_type: 'dirichlet', 'BDeu', or 'K2'
            string indicting which type of prior to use for the model parameters.
            - If 'prior_type' is 'dirichlet', the following must be provided:
                'pseudo_counts' = dirichlet hyperparameters; a dict containing, for each variable, a 2-D
                 array of the shape (node_card, product of parents_card) with a "virtual" count for each
                 variable state in the CPD, that is added to the state counts.
                 (lexicographic ordering of states assumed)
            - If 'prior_type' is 'BDeu', then an 'equivalent_sample_size'
                must be specified instead of 'pseudo_counts'. This is equivalent to
                'prior_type=dirichlet' and using uniform 'pseudo_counts' of
                `equivalent_sample_size/(node_cardinality*np.prod(parents_cardinalities))` for each node.
                'equivalent_sample_size' can either be a numerical value or a dict that specifies
                the size for each variable seperately.
            - A prior_type of 'K2' is a shorthand for 'dirichlet' + setting every pseudo_count to 1,
                regardless of the cardinality of the variable.

        Returns
        -------
        parameters: list
            List of TabularCPDs, one for each variable of the model

        Examples
        --------
        >>> import numpy as np
        >>> import pandas as pd
        >>> from pgmpy.models import BayesianModel
        >>> from pgmpy.estimators import BayesianEstimator
        >>> values = pd.DataFrame(np.random.randint(low=0, high=2, size=(1000, 4)),
        ...                       columns=['A', 'B', 'C', 'D'])
        >>> model = BayesianModel([('A', 'B'), ('C', 'B'), ('C', 'D')])
        >>> estimator = BayesianEstimator(model, values)
        >>> estimator.get_parameters(prior_type='BDeu', equivalent_sample_size=5)
        [<TabularCPD representing P(C:2) at 0x7f7b534251d0>,
        <TabularCPD representing P(B:2 | C:2, A:2) at 0x7f7b4dfd4da0>,
        <TabularCPD representing P(A:2) at 0x7f7b4dfd4fd0>,
        <TabularCPD representing P(D:2 | C:2) at 0x7f7b4df822b0>]
        """
        parameters = []
        for node in self.model.nodes():
            _equivalent_sample_size = (
                equivalent_sample_size[node]
                if isinstance(equivalent_sample_size, dict)
                else equivalent_sample_size
            )
            _pseudo_counts = pseudo_counts[node] if pseudo_counts else None

            cpd = self.estimate_cpd(
                node,
                prior_type=prior_type,
                equivalent_sample_size=_equivalent_sample_size,
                pseudo_counts=_pseudo_counts,
            )
            parameters.append(cpd)

        return parameters

    def estimate_cpd(
        self, node, prior_type="BDeu", pseudo_counts=[], equivalent_sample_size=5
    ):
        """
        Method to estimate the CPD for a given variable.

        Parameters
        ----------
        node: int, string (any hashable python object)
            The name of the variable for which the CPD is to be estimated.

        prior_type: 'dirichlet', 'BDeu', 'K2',
            string indicting which type of prior to use for the model parameters.
            - If 'prior_type' is 'dirichlet', the following must be provided:
                'pseudo_counts' = dirichlet hyperparameters; 2-D array of shape
                 (node_card, product of parents_card) with a "virtual" count for
                 each variable state in the CPD.
                 The virtual counts are added to the actual state counts found in the data.
                 (if a list is provided, a lexicographic ordering of states is assumed)
            - If 'prior_type' is 'BDeu', then an 'equivalent_sample_size'
                must be specified instead of 'pseudo_counts'. This is equivalent to
                'prior_type=dirichlet' and using uniform 'pseudo_counts' of
                `equivalent_sample_size/(node_cardinality*np.prod(parents_cardinalities))`.
            - A prior_type of 'K2' is a shorthand for 'dirichlet' + setting every pseudo_count to 1,
                regardless of the cardinality of the variable.

        Returns
        -------
        CPD: TabularCPD

        Examples
        --------
        >>> import pandas as pd
        >>> from pgmpy.models import BayesianModel
        >>> from pgmpy.estimators import BayesianEstimator
        >>> data = pd.DataFrame(data={'A': [0, 0, 1], 'B': [0, 1, 0], 'C': [1, 1, 0]})
        >>> model = BayesianModel([('A', 'C'), ('B', 'C')])
        >>> estimator = BayesianEstimator(model, data)
        >>> cpd_C = estimator.estimate_cpd('C', prior_type="dirichlet", pseudo_counts=[1, 2])
        >>> print(cpd_C)
        ╒══════╤══════╤══════╤══════╤════════════════════╕
        │ A    │ A(0) │ A(0) │ A(1) │ A(1)               │
        ├──────┼──────┼──────┼──────┼────────────────────┤
        │ B    │ B(0) │ B(1) │ B(0) │ B(1)               │
        ├──────┼──────┼──────┼──────┼────────────────────┤
        │ C(0) │ 0.25 │ 0.25 │ 0.5  │ 0.3333333333333333 │
        ├──────┼──────┼──────┼──────┼────────────────────┤
        │ C(1) │ 0.75 │ 0.75 │ 0.5  │ 0.6666666666666666 │
        ╘══════╧══════╧══════╧══════╧════════════════════╛

        """
        node_cardinality = len(self.state_names[node])
        parents = sorted(self.model.get_parents(node))
        parents_cardinalities = [len(self.state_names[parent]) for parent in parents]
        cpd_shape = (node_cardinality, np.prod(parents_cardinalities, dtype=int))

        if prior_type == "K2":
            pseudo_counts = np.ones(cpd_shape, dtype=int)
        elif prior_type == "BDeu":
            alpha = float(equivalent_sample_size) / (
                node_cardinality * np.prod(parents_cardinalities)
            )
            pseudo_counts = np.ones(cpd_shape, dtype=float) * alpha
        elif prior_type == "dirichlet":
            pseudo_counts = np.array(pseudo_counts)
            if pseudo_counts.shape != cpd_shape:
                raise ValueError(
                    "The shape of pseudo_counts for the node: {node} must be of shape: {shape}".format(
                        node=node, shape=str(cpd_shape)
                    )
                )
        else:
            raise ValueError("'prior_type' not specified")

        state_counts = self.state_counts(node)
        bayesian_counts = state_counts + pseudo_counts

        cpd = TabularCPD(
            node,
            node_cardinality,
            np.array(bayesian_counts),
            evidence=parents,
            evidence_card=parents_cardinalities,
            state_names={var: self.state_names[var] for var in chain([node], parents)},
        )
        cpd.normalize()
        return cpd



# -*- coding: utf-8 -*-

from __future__ import division

import numpy as np
import pandas as pd
from scipy.stats import multivariate_normal

from pgmpy.factors.base import BaseFactor


class LinearGaussianCPD(BaseFactor):
    r"""
    For, X -> Y the Linear Gaussian model assumes that the mean
    of Y is a linear function of mean of X and the variance of Y does
    not depend on X.

    For example,

    .. math::

      p(Y|X) = N(-2x + 0.9 ; 1)

    Here, :math:`x` is the mean of the variable :math:`X`.

    Let :math:`Y` be a continuous variable with continuous parents
    :math:`X1, X2, \cdots, Xk`. We say that :math:`Y` has a linear Gaussian CPD
    if there are parameters :math:`\beta_0, \beta_1, ..., \beta_k`
    and :math:`\sigma_2` such that,

    .. math:: p(Y |x1, x2, ..., xk) = \mathcal{N}(\beta_0 + x1*\beta_1 + ......... + xk*\beta_k ; \sigma_2)

    In vector notation,

    .. math:: p(Y |x) = \mathcal{N}(\beta_0 + \boldmath{β}.T * \boldmath{x} ; \sigma_2)

    References
    ----------
    .. [1] https://cedar.buffalo.edu/~srihari/CSE574/Chap8/Ch8-PGM-GaussianBNs/8.5%20GaussianBNs.pdf
    """

    def __init__(
        self, variable, evidence_mean, evidence_variance, evidence=[], beta=None
    ):
        """
        Parameters
        ----------

        variable: any hashable python object
            The variable whose CPD is defined.

        evidence_mean: Mean vector (numpy array) of the joint distribution, X

        evidence_variance: int, float
            The variance of the multivariate gaussian, X = ['x1', 'x2', ..., 'xn']

        evidence: iterable of any hashabale python objects
            An iterable of the parents of the variable. None if there are no parents.

        beta (optional): iterable of int or float
            An iterable representing the coefficient vector of the linear equation.
            The first term represents the constant term in the linear equation.


        Examples
        --------

        # For P(Y| X1, X2, X3) = N(-2x1 + 3x2 + 7x3 + 0.2; 9.6)

        >>> cpd = LinearGaussianCPD('Y',  [0.2, -2, 3, 7], 9.6, ['X1', 'X2', 'X3'])
        >>> cpd.variable
        'Y'
        >>> cpd.evidence
        ['x1', 'x2', 'x3']
        >>> cpd.beta_vector
        [0.2, -2, 3, 7]

        """
        self.variable = variable
        self.mean = evidence_mean
        self.variance = evidence_variance
        self.evidence = evidence
        self.sigma_yx = None

        variables = [variable] + evidence
        super(LinearGaussianCPD, self).__init__(
            variables, pdf="gaussian", mean=self.mean, covariance=self.variance
        )

    def sum_of_product(self, xi, xj):
        prod_xixj = xi * xj
        return np.sum(prod_xixj)

    def maximum_likelihood_estimator(self, data, states):
        """
        Fit using MLE method.

        Parameters
        ----------
        data: pandas.DataFrame or 2D array
            Dataframe of values containing samples from the conditional distribution, (Y|X)
            and corresponding X values.

        states: All the input states that are jointly gaussian.

        Returns
        -------
        beta, variance (tuple): Returns estimated betas and the variance.
        """
        x_df = pd.DataFrame(data, columns=states)
        x_len = len(self.evidence)

        sym_coefs = []
        for i in range(0, x_len):
            sym_coefs.append("b" + str(i + 1) + "_coef")

        sum_x = x_df.sum()
        x = [sum_x["(Y|X)"]]
        coef_matrix = pd.DataFrame(columns=sym_coefs)

        # First we compute just the coefficients of beta_1 to beta_N.
        # Later we compute beta_0 and append it.
        for i in range(0, x_len):
            x.append(self.sum_of_product(x_df["(Y|X)"], x_df[self.evidence[i]]))
            for j in range(0, x_len):
                coef_matrix.loc[i, sym_coefs[j]] = self.sum_of_product(
                    x_df[self.evidence[i]], x_df[self.evidence[j]]
                )

        coef_matrix.insert(0, "b0_coef", sum_x[self.evidence].values)
        row_1 = np.append([len(x_df)], sum_x[self.evidence].values)
        coef_matrix.loc[-1] = row_1
        coef_matrix.index = coef_matrix.index + 1  # shifting index
        coef_matrix.sort_index(inplace=True)

        beta_coef_matrix = np.matrix(coef_matrix.values, dtype="float")
        coef_inv = np.linalg.inv(beta_coef_matrix)
        beta_est = np.array(np.matmul(coef_inv, np.transpose(x)))
        self.beta = beta_est[0]

        sigma_est = 0
        x_len_df = len(x_df)
        for i in range(0, x_len):
            for j in range(0, x_len):
                sigma_est += (
                    self.beta[i + 1]
                    * self.beta[j + 1]
                    * (
                        self.sum_of_product(
                            x_df[self.evidence[i]], x_df[self.evidence[j]]
                        )
                        / x_len_df
                        - np.mean(x_df[self.evidence[i]])
                        * np.mean(x_df[self.evidence[j]])
                    )
                )

        sigma_est = np.sqrt(
            self.sum_of_product(x_df["(Y|X)"], x_df["(Y|X)"]) / x_len_df
            - np.mean(x_df["(Y|X)"]) * np.mean(x_df["(Y|X)"])
            - sigma_est
        )
        self.sigma_yx = sigma_est
        return self.beta, self.sigma_yx

    def fit(self, data, states, estimator=None, complete_samples_only=True, **kwargs):
        """
        Determine βs from data

        Parameters
        ----------
        data: pandas.DataFrame
            Dataframe containing samples from the conditional distribution, p(Y|X)
            estimator: 'MLE' or 'MAP'

        completely_samples_only: boolean (True or False)
            Are they downsampled or complete? Defaults to True

        """
        if estimator == "MLE":
            mean, variance = self.maximum_likelihood_estimator(data, states)
        elif estimator == "MAP":
            raise NotImplementedError(
                "fit method has not been implemented using Maximum A-Priori (MAP)"
            )

        return mean, variance

    @property
    def pdf(self):
        def _pdf(*args):
            # The first element of args is the value of the variable on which CPD is defined
            # and the rest of the elements give the mean values of the parent
            # variables.
            mean = (
                sum([arg * coeff for (arg, coeff) in zip(args[1:], self.mean)])
                + self.mean[0]
            )
            return multivariate_normal.pdf(
                args[0], np.array(mean), np.array([[self.variance]])
            )

        return _pdf

    def copy(self):
        """
        Returns a copy of the distribution.

        Returns
        -------
        LinearGaussianCPD: copy of the distribution

        Examples
        --------
        >>> from pgmpy.factors.continuous import LinearGaussianCPD
        >>> cpd = LinearGaussianCPD('Y',  [0.2, -2, 3, 7], 9.6, ['X1', 'X2', 'X3'])
        >>> copy_cpd = cpd.copy()
        >>> copy_cpd.variable
        'Y'
        >>> copy_cpd.evidence
        ['X1', 'X2', 'X3']
        """
        copy_cpd = LinearGaussianCPD(
            self.variable, self.beta, self.variance, list(self.evidence)
        )

        return copy_cpd

    def __str__(self):
        if self.evidence and list(self.mean):
            # P(Y| X1, X2, X3) = N(-2*X1_mu + 3*X2_mu + 7*X3_mu; 0.2)
            rep_str = "P({node} | {parents}) = N({mu} + {b_0}; {sigma})".format(
                node=str(self.variable),
                parents=", ".join([str(var) for var in self.evidence]),
                mu=" + ".join(
                    [
                        "{coeff}*{parent}".format(coeff=coeff, parent=parent)
                        for coeff, parent in zip(self.mean, self.evidence)
                    ]
                ),
                b_0=str(self.mean[0]),
                sigma=str(self.variance),
            )
        else:
            # P(X) = N(1, 4)
            rep_str = "P({X}) = N({beta_0}; {variance})".format(
                X=str(self.variable),
                beta_0=str(self.mean[0]),
                variance=str(self.variance),
            )
        return rep_str



import unittest

import pandas as pd
import numpy as np

from pgmpy.models import SEMGraph, SEM
from pgmpy.estimators import SEMEstimator, IVEstimator


class TestSEMEstimator(unittest.TestCase):
    def setUp(self):
        self.custom = SEMGraph(
            ebunch=[("a", "b"), ("b", "c")], latents=[], err_corr=[], err_var={}
        )
        a = np.random.randn(10 ** 3)
        b = a + np.random.normal(loc=0, scale=0.1, size=10 ** 3)
        c = b + np.random.normal(loc=0, scale=0.2, size=10 ** 3)
        self.custom_data = pd.DataFrame({"a": a, "b": b, "c": c})
        self.custom_data -= self.custom_data.mean(axis=0)
        self.custom_lisrel = self.custom.to_lisrel()

        self.demo = SEMGraph(
            ebunch=[
                ("xi1", "x1"),
                ("xi1", "x2"),
                ("xi1", "x3"),
                ("xi1", "eta1"),
                ("eta1", "y1"),
                ("eta1", "y2"),
                ("eta1", "y3"),
                ("eta1", "y4"),
                ("eta1", "eta2"),
                ("xi1", "eta2"),
                ("eta2", "y5"),
                ("eta2", "y6"),
                ("eta2", "y7"),
                ("eta2", "y8"),
            ],
            latents=["xi1", "eta1", "eta2"],
            err_corr=[
                ("y1", "y5"),
                ("y2", "y6"),
                ("y3", "y7"),
                ("y4", "y8"),
                ("y2", "y4"),
                ("y6", "y8"),
            ],
            err_var={},
        )
        self.demo_lisrel = self.demo.to_lisrel()

        self.demo_data = pd.read_csv(
            "pgmpy/tests/test_estimators/testdata/democracy1989a.csv",
            index_col=0,
            header=0,
        )

        self.union = SEMGraph(
            ebunch=[
                ("yrsmill", "unionsen"),
                ("age", "laboract"),
                ("age", "deferenc"),
                ("deferenc", "laboract"),
                ("deferenc", "unionsen"),
                ("laboract", "unionsen"),
            ],
            latents=[],
            err_corr=[("yrsmill", "age")],
            err_var={},
        )
        self.union_lisrel = self.union.to_lisrel()

        self.union_data = pd.read_csv(
            "pgmpy/tests/test_estimators/testdata/union1989b.csv", index_col=0, header=0
        )

    def test_get_init_values(self):
        demo_estimator = SEMEstimator(self.demo)
        for method in ["random", "std"]:
            B_init, zeta_init = demo_estimator.get_init_values(
                data=self.demo_data, method=method
            )

            demo_lisrel = self.demo.to_lisrel()
            m = len(demo_lisrel.eta)
            self.assertEqual(B_init.shape, (m, m))
            self.assertEqual(zeta_init.shape, (m, m))

            union_estimator = SEMEstimator(self.union)
            B_init, zeta_init = union_estimator.get_init_values(
                data=self.union_data, method=method
            )
            union_lisrel = self.union.to_lisrel()
            m = len(union_lisrel.eta)
            self.assertEqual(B_init.shape, (m, m))
            self.assertEqual(zeta_init.shape, (m, m))

    @unittest.skip
    def test_demo_estimator_random_init(self):
        estimator = SEMEstimator(self.demo)
        summary = estimator.fit(self.demo_data, method="ml")

    @unittest.skip
    def test_union_estimator_random_init(self):
        estimator = SEMEstimator(self.union_lisrel)
        summary = estimator.fit(
            self.union_data, method="ml", opt="adam", max_iter=10 ** 6, exit_delta=1e-1
        )

    @unittest.skip
    def test_custom_estimator_random_init(self):
        estimator = SEMEstimator(self.custom_lisrel)
        summary = estimator.fit(
            self.custom_data, method="ml", max_iter=10 ** 6, opt="adam"
        )
        summary = estimator.fit(
            self.custom_data, method="uls", max_iter=10 ** 6, opt="adam"
        )
        summary = estimator.fit(
            self.custom_data,
            method="gls",
            max_iter=10 ** 6,
            opt="adam",
            W=np.ones((3, 3)),
        )

    @unittest.skip
    def test_union_estimator_std_init(self):
        estimator = SEMEstimator(self.union_lisrel)
        summary = estimator.fit(
            self.union_data,
            method="ml",
            opt="adam",
            init_values="std",
            max_iter=10 ** 6,
            exit_delta=1e-1,
        )

    @unittest.skip
    def test_custom_estimator_std_init(self):
        estimator = SEMEstimator(self.custom_lisrel)
        summary = estimator.fit(
            self.custom_data,
            method="ml",
            init_values="std",
            max_iter=10 ** 6,
            opt="adam",
        )


class TestIVEstimator(unittest.TestCase):
    def setUp(self):
        self.model = SEM.from_graph(
            ebunch=[
                ("Z1", "X", 1.0),
                ("Z2", "X", 1.0),
                ("Z2", "W", 1.0),
                ("W", "U", 1.0),
                ("U", "X", 1.0),
                ("U", "Y", 1.0),
                ("X", "Y", 1.0),
            ],
            latents=["U"],
            err_var={"Z1": 1, "Z2": 1, "W": 1, "X": 1, "U": 1, "Y": 1},
        )
        self.generated_data = self.model.to_lisrel().generate_samples(100000)

    def test_fit(self):
        estimator = IVEstimator(self.model)
        param, summary = estimator.fit(X="X", Y="Y", data=self.generated_data)
        self.assertTrue((param - 1) < 0.01)



# -*- coding: UTF-8 -*-
"""
For the student example the ProbModelXML file should be:

<?xml version=“1.0” encoding=“UTF-8”?>
<ProbModelXML formatVersion=“1.0”>
    <ProbNet type="BayesianNetwork">
        <AdditionalConstraints />
        <Comment>
            Student example model from Probabilistic Graphical Models:
            Principles and Techniques by Daphne Koller
        </Comment>
        <Language>
            English
        </Language>
        <AdditionalProperties />
        <Variables>
            <Variable name="intelligence" type="FiniteState" role="Chance">
                <Comment />
                <Coordinates />
                <AdditionalProperties />
                <States>
                    <State name="smart"><AdditionalProperties /></State>
                    <State name="dumb"><AdditionalProperties /></State>
                </States>
            </Variable>
            <Variable name="difficulty" type="FiniteState" role="Chance">
                <Comment />
                <Coordinates />
                <AdditionalProperties />
                <States>
                    <State name="difficult"><AdditionalProperties /></State>
                    <State name="easy"><AdditionalProperties /></State>
                </States>
            </Variable>
            <Variable name="grade" type="FiniteState" role="Chance">
                <Comment />
                <Coordinates />
                <AdditionalProperties />
                <States>
                    <State name="grade_A"><AdditionalProperties /></State>
                    <State name="grade_B"><AdditionalProperties /></State>
                    <State name="grade_C"><AdditionalProperties /></State>
                </States>
            </Variable>
            <Variable name="recommendation_letter" type="FiniteState"
                    role="Chance">
                <Comment />
                <Coordinates />
                <AdditionalProperties />
                <States>
                    <State name="good"><AdditionalProperties /></State>
                    <State name="bad"><AdditionalProperties /></State>
                </States>
            </Variable>
            <Variable name="SAT" type="FiniteState" role="Chance">
                <Comment />
                <Coordinates />
                <AdditionalProperties />
                <States>
                    <State name="high"><AdditionalProperties /></State>
                    <State name="low"><AdditionalProperties /></State>
                </States>
            </Variable>
        </Variables>
        <Links>
            <Link var1="difficulty" var2="grade" directed=1>
                <Comment>Directed Edge from difficulty to grade</Comment>
                <Label>diff_to_grad</Label>
                <AdditionalProperties />
            </Link>
            <Link var1="intelligence" var2="grade" directed=1>
                <Comment>Directed Edge from intelligence to grade</Comment>
                <Label>intel_to_grad</Label>
                <AdditionalProperties />
            </Link>
            <Link var1="intelligence" var2="SAT" directed=1>
                <Comment>Directed Edge from intelligence to SAT</Comment>
                <Label>intel_to_sat</Label>
                <AdditionalProperties />
            </Link>
            <Link var1="grade" var2="recommendation_letter" directed=1>
                <Comment>Directed Edge from grade to
                    recommendation_letter</Comment>
                <Label>grad_to_reco</Label>
                <AdditionalProperties />
            </Link>
        </Links>
        <Potential type="Table" role="ConditionalProbability" label=string>
            <Comment>CPDs in the form of table</Comment>
            <AdditionalProperties />
            <!--
                There is no specification in the paper about
                how the tables should be represented.
            -->
        </Potential>
    </ProbNet>
    <Policies />
    <InferenceOptions />
    <Evidence>
        <EvidenceCase>
            <Finding variable=string state=string stateIndex=integer
                numericValue=number/>
        </EvidenceCase>
    </Evidence>
</ProbModelXML>
"""

import warnings

try:
    from lxml import etree
except ImportError:
    try:
        import xml.etree.ElementTree as etree
    except ImportError:
        # import xml.etree.cElementTree as etree
        # print("running with cElementTree on Python 2.5+")
        # Commented out because behaviour is different from expected

        warnings.warn("Failed to import ElementTree from any known place")

import networkx as nx
import numpy as np

from pgmpy.models import BayesianModel
from pgmpy.factors.discrete import TabularCPD

# warnings.warn("Not Complete. Please use only for "
#               "reading and writing Bayesian Models.")


def generate_probmodelxml(model, encoding="utf-8", prettyprint=True):
    """
    Generate ProbModelXML lines for model.

    Parameters
    ----------
    model : Graph
        The Bayesian or Markov Model
    encoding : string (optional)
        Encoding for text data
    prettyprint: bool (optional)
        If True uses line breaks and indenting in output XML.

    Examples
    --------
    >>> G = nx.path_graph(5)
    >>> s = pgmpy.readwrite.generate_ProbModelXML(G) # doctest: +SKIP
    >>> for line in pgmpy.readwrite.generate_ProbModelXML(G):  #doctest: +SKIP
    ...     print(line)
    """
    writer = ProbModelXMLWriter(G, encoding=encoding, prettyprint=prettyprint)
    for line in str(writer).splitlines():
        yield line


# @open_file(1, mode='wb')
def write_probmodelxml(model, path, encoding="utf-8", prettyprint=True):
    """
    Write model in ProbModelXML format to path.

    Parameters
    ----------
    model : A NetworkX graph
            Bayesian network or Markov network
    path : file or string
            File or filename to write.
            Filenames ending in .gz or .bz2 will be compressed.
    encoding : string (optional)
            Encoding for text data.
    prettyprint : bool (optional)
            If True use line breaks and indenting in output XML.

    Examples
    --------
    >>> G = nx.path_graph(4)
    >>> pgmpy.readwrite.write_probmodelxml(G, "test.probmodelxml")
    """
    writer = ProbModelXMLWriter(model, path, encoding=encoding, prettyprint=prettyprint)
    writer.dump(path)


# @open_file(0, mode='rb')
def read_probmodelxml(path):
    """
    Read model in ProbModelXML format from path.

    Parameters
    ----------
    path : file or string
        file or filename from which to read.

    Returns
    -------
    model : NetworkX Graph
            A BayesianModel or MarkovModel object depending on the
            type of model the XML represents.

    Examples
    --------
    >>> G = pgmpy.readwrite.read_probmodelxml('test.probModelXML')
    """
    reader = ProbModelXMLReader(path=path)
    return reader.make_network()


def parse_probmodelxml(string):
    """
    Read model in ProbModelXML format from string.

    Parameters
    ----------
    string : string
        String containing ProbModelXML information.
        (e.g., contents of a ProbModelXML file).

    Returns
    -------
    model : NetworkX Graph
        A BayesianModel or MarkovModel object depending on the XML given.

    Examples
    --------
    >>> G = nx.path_graph(4)
    >>> linefeed = chr(10)
    >>> s = linefeed.join(pgmpy.readwrite.generate_probmodelxml(G))
    >>> H = pgmpy.readwrite.parse_probmodelxml(s)
    """
    reader = ProbModelXMLReader(string=string)
    return reader.make_network()


def get_probmodel_data(model):
    """
    Returns the model_data based on the given model.

    Parameters
    ----------
    model: BayesianModel instance
        Model to write

    Return
    ------
    model_data: dict
        dictionary containing model data of the given model.

    Examples
    --------
    >>> model_data = pgmpy.readwrite.get_model_data(model)
    >>> writer.get_model_data(model)
    """
    if not isinstance(model, BayesianModel):
        raise TypeError("Model must an instance of BayesianModel.")
    model_data = {"probnet": {"type": "BayesianNetwork", "Variables": {}}}

    variables = model.nodes()
    for var in variables:
        model_data["probnet"]["Variables"][var] = model.nodes[var]

    model_data["probnet"]["edges"] = {}
    edges = model.edges()
    for edge in edges:
        model_data["probnet"]["edges"][str(edge)] = model.adj[edge[0]][edge[1]]

    model_data["probnet"]["Potentials"] = []
    cpds = model.get_cpds()
    for cpd in cpds:
        potential_dict = {}
        potential_dict["Variables"] = {}
        evidence = cpd.variables[:0:-1]
        if evidence:
            potential_dict["Variables"][cpd.variable] = evidence
        else:
            potential_dict["Variables"][cpd.variable] = []
        potential_dict["type"] = "Table"
        potential_dict["role"] = "conditionalProbability"
        potential_dict["Values"] = (
            " ".join([str(val) for val in cpd.values.ravel().astype(float)]) + " "
        )
        model_data["probnet"]["Potentials"].append(potential_dict)

    return model_data


class ProbModelXMLWriter(object):
    """
    Class for writing models in ProbModelXML format.
    """

    def __init__(self, model_data, encoding="utf-8", prettyprint=True):
        """
        Initialize a ProbModelXMLWriter Object.

        Parameters
        ----------
        model : A BayesianModel or MarkovModel
            The model to write.
        encoding : string (optional)
            Encoding for text data
        prettyprint : bool (optional)
            If True uses line breaks and indentation in output XML.

        Examples
        --------

        Refernces
        ---------
        [1] http://leo.ugr.es/pgm2012/submissions/pgm2012_submission_43.pdf
        [2] http://www.cisiad.uned.es/techreports/ProbModelXML.pdf
        """
        # TODO: add policies, InferenceOptions, Evidence
        # TODO: add parsing of language and comments and additional properties
        self.data = model_data
        self.encoding = encoding
        self.prettyprint = prettyprint

        # Creating initial tags
        self.xml = etree.Element("ProbModelXML", attrib={"formatVersion": "1.0"})
        self.probnet = etree.SubElement(self.xml, "ProbNet")
        self.variables = etree.SubElement(self.probnet, "Variables")
        self.links = etree.SubElement(self.probnet, "Links")
        self.potentials = etree.SubElement(self.probnet, "Potentials")
        self.additional_constraints = etree.SubElement(
            self.probnet, "AdditionalConstraints"
        )

        # adding information for probnet
        self.probnet.attrib["type"] = self.data["probnet"]["type"]
        try:
            etree.SubElement(self.probnet, "Language").text = self.data["probnet"][
                "Language"
            ]
        except KeyError:
            pass
        try:
            etree.SubElement(self.probnet, "Comment").text = self.data["probnet"][
                "Comment"
            ]
        except KeyError:
            pass
        try:
            self._add_additional_properties(
                self.xml, self.data["probnet"]["AdditionalProperties"]
            )
        except KeyError:
            etree.SubElement(self.probnet, "AdditionalProperties")
        try:
            self._add_decision_criteria(self.data["probnet"]["DecisionCriteria"])
        except KeyError:
            etree.SubElement(self.probnet, "DecisionCriteria")

        # Add Additional Constraints
        if "AdditionalConstraints" in self.data["probnet"]:
            for constraint in sorted(self.data["probnet"]["AdditionalConstraints"]):
                self._add_constraint(constraint)

        # Add variables
        for variable in sorted(self.data["probnet"]["Variables"]):
            self._add_variable(variable)

        # Add edges
        for edge in sorted(self.data["probnet"]["edges"]):
            self._add_link(edge)

        # Add Potentials
        for potential in self.data["probnet"]["Potentials"]:
            self._add_potential(potential, self.potentials)

    def __str__(self):
        """
        Return the XML as string.
        """
        if self.prettyprint:
            self.indent(self.xml)
        return etree.tostring(self.xml, encoding=self.encoding)

    @staticmethod
    def _add_additional_properties(position, properties_dict):
        """
        Sets AdditionalProperties of the ProbModelXML.
        """
        add_prop = etree.SubElement(position, "AdditionalProperties")
        for key, value in properties_dict.items():
            etree.SubElement(add_prop, "Property", attrib={"name": key, "value": value})

    def _add_variable(self, variable):
        """
        Adds a node to the ProbModelXML.
        """
        # TODO: Add feature for accepting additional properties of states.
        variable_data = self.data["probnet"]["Variables"][variable]
        variable_element = etree.SubElement(
            self.variables,
            "Variable",
            attrib={
                "name": variable,
                "type": variable_data["type"],
                "role": variable_data["role"],
            },
        )

        try:
            etree.SubElement(variable_element, "Comment").text = variable_data[
                "Comment"
            ]
        except KeyError:
            pass
        try:
            etree.SubElement(
                variable_element, "Coordinates", variable_data["Coordinates"]
            )
        except KeyError:
            pass

        try:
            for key, value in sorted(variable_data["AdditionalProperties"].items()):
                etree.SubElement(
                    variable_element, "Property", attrib={"name": key, "value": value}
                )
        except KeyError:
            etree.SubElement(variable_element, "AdditionalProperties")
        states = etree.SubElement(variable_element, "States")
        for s in sorted(variable_data["States"]):
            state = etree.SubElement(states, "State", attrib={"name": s})
            try:
                self._add_additional_properties(
                    state, variable_data["States"][s]["AdditionalProperties"]
                )
            except KeyError:
                etree.SubElement(state, "AdditionalProperties")

    def _add_link(self, edge):
        """
        Adds an edge to the ProbModelXML.
        """
        edge_data = self.data["probnet"]["edges"][edge]
        if isinstance(edge, str):
            edge = eval(edge)
        link = etree.SubElement(
            self.links,
            "Link",
            attrib={
                "var1": edge[0],
                "var2": edge[1],
                "directed": edge_data["directed"],
            },
        )
        try:
            etree.SubElement(link, "Comment").text = edge_data["Comment"]
        except KeyError:
            pass
        try:
            etree.SubElement(link, "Label").text = edge_data["Label"]
        except KeyError:
            pass
        try:
            self._add_additional_properties(link, edge_data["AdditionalProperties"])
        except KeyError:
            etree.SubElement(link, "AdditionalProperties")

    def _add_constraint(self, constraint):
        """
        Adds constraint to the ProbModelXML.
        """
        constraint_data = self.data["probnet"]["AdditionalConstraints"][constraint]
        constraint_element = etree.SubElement(
            self.additional_constraints, "Constraint", attrib={"name": constraint}
        )
        for argument in sorted(constraint_data):
            name = argument
            value = constraint_data[name]
            etree.SubElement(
                constraint_element, "Argument", attrib={"name": name, "value": value}
            )

    def _add_decision_criteria(self, criteria_dict):
        """
        Adds Decision Criteria to the ProbModelXML.

        Parameters
        ----------
        criteria_dict: dict
            Dictionary containing Deecision Criteria data.
            For example: {'effectiveness': {}, 'cost': {}}

        Examples
        -------
        >>> writer = ProbModelXMLWriter(model)
        >>> writer._add_decision_criteria(criteria_dict)
        """
        decision_tag = etree.SubElement(self.xml, "DecisionCriteria", attrib={})
        for criteria in sorted(criteria_dict):
            criteria_tag = etree.SubElement(
                decision_tag, "Criterion", attrib={"name": criteria}
            )
            self._add_additional_properties(criteria_tag, criteria_dict[criteria])

    def _add_potential(self, potential, parent_tag):
        """
        Adds Potentials to the ProbModelXML.

        Parameters
        ----------
        potential: dict
            Dictionary containing Potential data.
            For example: {'role': 'Utility',
                          'Variables': ['D0', 'D1', 'C0', 'C1'],
                          'type': 'Tree/ADD',
                          'UtilityVaribale': 'U1'}
        parent_tag: etree Element
            etree element which would contain potential tag
            For example: <Element Potentials at 0x7f315fc44b08>
                         <Element Branch at 0x7f315fc44c88>
                         <Element Branch at 0x7f315fc44d88>
                         <Element Subpotentials at 0x7f315fc44e48>

        Examples
        --------
        >>> writer = ProbModelXMLWriter(model)
        >>> writer._add_potential(potential, parent_tag)
        """
        potential_type = potential["type"]
        try:
            potential_tag = etree.SubElement(
                parent_tag,
                "Potential",
                attrib={"type": potential["type"], "role": potential["role"]},
            )
        except KeyError:
            potential_tag = etree.SubElement(
                parent_tag, "Potential", attrib={"type": potential["type"]}
            )
        self._add_element(potential, "Comment", potential_tag)
        if "AdditionalProperties" in potential:
            self._add_additional_properties(
                potential_tag, potential["AdditionalProperties"]
            )
        if potential_type == "delta":
            etree.SubElement(
                potential_tag, "Variable", attrib={"name": potential["Variable"]}
            )
            self._add_element(potential, "State", potential_tag)
            self._add_element(potential, "StateIndex", potential_tag)
            self._add_element(potential, "NumericValue", potential_tag)
        else:
            if "UtilityVariable" in potential:
                etree.SubElement(
                    potential_tag,
                    "UtilityVariable",
                    attrib={"name": potential["UtilityVariable"]},
                )
            if "Variables" in potential:
                variable_tag = etree.SubElement(potential_tag, "Variables")
                for var in sorted(potential["Variables"]):
                    etree.SubElement(variable_tag, "Variable", attrib={"name": var})
                    for child in sorted(potential["Variables"][var]):
                        etree.SubElement(
                            variable_tag, "Variable", attrib={"name": child}
                        )
            self._add_element(potential, "Values", potential_tag)
            if "UncertainValues" in potential:
                value_tag = etree.SubElement(
                    potential_tag, "UncertainValues", attrib={}
                )
                for value in sorted(potential["UncertainValues"]):
                    try:
                        etree.SubElement(
                            value_tag,
                            "Value",
                            attrib={
                                "distribution": value["distribution"],
                                "name": value["name"],
                            },
                        ).text = value["value"]
                    except KeyError:
                        etree.SubElement(
                            value_tag,
                            "Value",
                            attrib={"distribution": value["distribution"]},
                        ).text = value["value"]
            if "TopVariable" in potential:
                etree.SubElement(
                    potential_tag,
                    "TopVariable",
                    attrib={"name": potential["TopVariable"]},
                )
            if "Branches" in potential:
                branches_tag = etree.SubElement(potential_tag, "Branches")
                for branch in potential["Branches"]:
                    branch_tag = etree.SubElement(branches_tag, "Branch")
                    if "States" in branch:
                        states_tag = etree.SubElement(branch_tag, "States")
                        for state in sorted(branch["States"]):
                            etree.SubElement(
                                states_tag, "State", attrib={"name": state["name"]}
                            )
                    if "Potential" in branch:
                        self._add_potential(branch["Potential"], branch_tag)
                    self._add_element(potential, "Label", potential_tag)
                    self._add_element(potential, "Reference", potential_tag)
                    if "Thresholds" in branch:
                        thresholds_tag = etree.SubElement(branch_tag, "Thresholds")
                        for threshold in branch["Thresholds"]:
                            try:
                                etree.SubElement(
                                    thresholds_tag,
                                    "Threshold",
                                    attrib={
                                        "value": threshold["value"],
                                        "belongsTo": threshold["belongsTo"],
                                    },
                                )
                            except KeyError:
                                etree.SubElement(
                                    thresholds_tag,
                                    "Threshold",
                                    attrib={"value": threshold["value"]},
                                )
            self._add_element(potential, "Model", potential_tag)
            self._add_element(potential, "Coefficients", potential_tag)
            self._add_element(potential, "CovarianceMatrix", potential_tag)
            if "Subpotentials" in potential:
                subpotentials = etree.SubElement(potential_tag, "Subpotentials")
                for subpotential in potential["Subpotentials"]:
                    self._add_potential(subpotential, subpotentials)
            if "Potential" in potential:
                self._add_potential(potential["Potential"], potential_tag)
            if "NumericVariables" in potential:
                numvar_tag = etree.SubElement(potential_tag, "NumericVariables")
                for var in sorted(potential["NumericVariables"]):
                    etree.SubElement(numvar_tag, "Variable", attrib={"name": var})

    @staticmethod
    def _add_element(potential, var, potential_tag):
        """
        Helper function to add variable tag to the potential_tag

        Parameters
        ----------
        potential: dict
            Dictionary containing Potential data.
            For example: {'role': 'Utility',
                          'Variables': ['D0', 'D1', 'C0', 'C1'],
                          'type': 'Tree/ADD',
                          'UtilityVaribale': 'U1'}
        var: string
            New Element tag which needs to be added to the potential tag.
            For example: 'type'
        potential_tag: etree Element
            etree element which would contain potential tag
            For example: <Element Potentials at 0x7f315fc44b08>
                         <Element Branch at 0x7f315fc44c88>
                         <Element Branch at 0x7f315fc44d88>
                         <Element Subpotentials at 0x7f315fc44e48>

        Examples
        --------
        >>> writer = ProbModelXMLWriter(model)
        >>> writer._add_element(potential, 'State', parent_tag)
        """
        if var in potential:
            etree.SubElement(potential_tag, var).text = potential[var]

    def dump(self, stream):
        """
        Dumps the data to stream after appending header.
        """
        if self.prettyprint:
            self.indent(self.xml)
        document = etree.ElementTree(self.xml)
        header = '<?xml version="1.0" encoding="%s"?>' % self.encoding
        stream.write(header.encode(self.encoding))
        document.write(stream, encoding=self.encoding)

    def indent(self, elem, level=0):
        """
        Inplace prettyprint formatter.
        """
        i = "\n" + level * "  "
        if len(elem):
            if not elem.text or not elem.text.strip():
                elem.text = i + "  "
            if not elem.tail or not elem.tail.strip():
                elem.tail = i
            for elem in elem:
                self.indent(elem, level + 1)
            if not elem.tail or not elem.tail.strip():
                elem.tail = i
        else:
            if level and (not elem.tail or not elem.tail.strip()):
                elem.tail = i

    def write_file(self, filename):
        """
        Write the xml data into the file.

        Parameters
        ----------
        filename: Name of the file.

        Examples
        -------
        >>> writer = ProbModelXMLWriter(model)
        >>> writer.write_file(test_file)
        """
        writer = self.__str__()[:-1].decode("utf-8")
        with open(filename, "w") as fout:
            fout.write(writer)


class ProbModelXMLReader(object):
    """
    Class for reading ProbModelXML format from files or strings.
    """

    # TODO: add methods to parse policies, inferenceoption, evidence etc.
    # TODO: add reading formatVersion
    def __init__(self, path=None, string=None):
        """
        Initialize an instance of ProbModelXMLReader class.

        Parameters
        ----------
        path : file or string
            File containing ProbModelXML information.
        string : string
            String containing ProbModelXML information.

        Example
        -------
        >>> reader = ProbModelXMLReader('test.ProbModelXML')

        Structure of Probnet Object
        ---------------------------
        { probnet: { type:
                     Comment:
                     Language:
                     AdditionalProperties: { property_name1: property_value1,
                                           property_name2: property_value2,
                                                        ....
                                         }
                     Variables: { variable_name1: { type:
                                                    roles:
                                                    Comment:
                                                    Coordinates:
                                                    AdditionalProperties: { property_name1: property_value1,
                                                                            property_name2: property_value2,
                                                                                        ....
                                                                          }
                                                    states: { state1: {AdditionalProperties: {
                                                                                               ....
                                                                                               ....
                                                                                             }
                                                              state2: {AdditionalProperties: {
                                                                                               ....
                                                                                               ....
                                                                                             }
                                                                 .....
                                                            }
                                                  }
                                  variable_name2: {
                                                    ...........
                                                  }
                                      .........
                                 }
                     edges: { (var1, var2): { directed:
                                              Comment:
                                              Label:
                                              AdditionalProperties: { property_name1: property_value1,
                                                                      property_name2: property_value2,
                                                                                 .....
                                                                    }
                              (var3, var4): {
                                              .....
                                              .....
                                            }
                                   ........
                            }
                    }
        }

        References
        ----------
        [1] http://leo.ugr.es/pgm2012/submissions/pgm2012_submission_43.pdf
        [2] http://www.cisiad.uned.es/techreports/ProbModelXML.pdf
        """
        if path is not None:
            self.xml = etree.ElementTree(file=path)
        elif string is not None:
            self.xml = etree.fromstring(string)
        else:
            raise ValueError("Must specify either 'path' or 'string' as kwarg.")

        self.create_probnet()

    def create_probnet(self):
        """
        Returns a BayesianModel or MarkovModel object depending on the
        type of ProbModelXML passed to ProbModelXMLReader class.
        """
        self.probnet = {}
        # Add general properties
        probnet_elem = self.xml.find("ProbNet")
        self.probnet["type"] = probnet_elem.attrib["type"]
        if probnet_elem.find("Comment") is not None:
            self.add_comment(probnet_elem.find("Comment").text)
        if probnet_elem.find("Language") is not None:
            self.add_language(probnet_elem.find("Language").text)
        if probnet_elem.find("AdditionalProperties") is not None:
            self.probnet["AdditionalProperties"] = {}
            for prop in probnet_elem.find("AdditionalProperties"):
                self.add_additional_property(self.probnet["AdditionalProperties"], prop)

        # Add additional Constraints
        self.probnet["AdditionalConstraints"] = {}
        for constraint in probnet_elem.findall("AdditionalConstraints/Constraint"):
            self.add_probnet_additionalconstraints(constraint)

        # Add Decision Criterion
        self.probnet["DecisionCriteria"] = {}
        for criterion in probnet_elem.findall("DecisionCriteria/Criterion"):
            self.add_criterion(criterion)

        # Add nodes
        self.probnet["Variables"] = {}
        for variable in probnet_elem.find("Variables"):
            self.add_node(variable)

        # Add edges
        self.probnet["edges"] = {}
        for edge in probnet_elem.findall("Links/Link"):
            self.add_edge(edge)

        # Add CPD
        self.probnet["Potentials"] = []
        for potential in probnet_elem.findall("Potentials/Potential"):
            probnet_dict = {}
            self.add_potential(potential, probnet_dict)
            self.probnet["Potentials"].append(probnet_dict)

    def add_probnet_additionalconstraints(self, constraint):
        """
        Adds Additional Constraints to the probnet dict.

        Parameters
        ----------
        criterion: <Element Constraint at AdditionalConstraints Node in XML>
            etree Element consisting Constraint tag.

        Examples
        -------
        >>> reader = ProbModelXMLReader()
        >>> reader.add_additionalconstraints(constraint)
        """
        constraint_name = constraint.attrib["name"]
        self.probnet["AdditionalConstraints"][constraint_name] = {}
        for argument in constraint.findall("Argument"):
            argument_name = argument.attrib["name"]
            argument_value = argument.attrib["value"]
            self.probnet["AdditionalConstraints"][constraint_name][
                argument_name
            ] = argument_value

    def add_criterion(self, criterion):
        """
        Adds Decision Criteria to the probnet dict.

        Parameters
        ----------
        criterion: <Element Criterion at Decision Criteria Node in XML>
            etree Element consisting DecisionCritera tag.

        Examples
        -------
        >>> reader = ProbModelXMLReader()
        >>> reader.add_criterion(criterion)
        """
        criterion_name = criterion.attrib["name"]
        self.probnet["DecisionCriteria"][criterion_name] = {}
        if criterion.find("AdditionalProperties/Property") is not None:
            for prop in criterion.findall("AdditionalProperties/Property"):
                prop_name = prop.attrib["name"]
                prop_value = prop.attrib["value"]
                self.probnet["DecisionCriteria"][criterion_name][
                    "AdditionalProperties"
                ][prop_name] = prop_value

    def add_comment(self, comment):
        """
        Adds Comment to the probnet dict.

        Parameters
        ----------
        comment: string
            String consisting of comment.

        Examples
        -------
        >>> reader = ProbModelXMLReader()
        >>> reader.add_comment(comment)
        """
        self.probnet["Comment"] = comment

    def add_language(self, language):
        """
        Adds Language to the probnet dict.

        Parameters
        ----------
        comment: string
            String consisting of language.

        Examples
        -------
        >>> reader = ProbModelXMLReader()
        >>> reader.add_language(language)
        """
        self.probnet["Language"] = language

    @staticmethod
    def add_additional_property(place, prop):
        place[prop.attrib["name"]] = prop.attrib["value"]

    def add_node(self, variable):
        """
        Adds Variables to the probnet dict.

        Parameters
        ----------
        variable: <Element Variable at Variables Node in XML>
            etree Element consisting Variable tag.

        Examples
        -------
        >>> reader = ProbModelXMLReader()
        >>> reader.add_node(variable)
        """
        # TODO: Do some checks with variable type and roles. Right now I don't know when they are to be used.
        variable_name = variable.attrib["name"]
        self.probnet["Variables"][variable_name] = {}
        self.probnet["Variables"][variable_name]["type"] = variable.attrib["type"]
        self.probnet["Variables"][variable_name]["role"] = variable.attrib["role"]
        if variable.find("Comment") is not None:
            self.probnet["Variables"][variable_name]["Comment"] = variable.find(
                "Comment"
            ).text
        if variable.find("Coordinates") is not None:
            self.probnet["Variables"][variable_name]["Coordinates"] = variable.find(
                "Coordinates"
            ).attrib
        if variable.find("AdditionalProperties/Property") is not None:
            self.probnet["Variables"][variable_name]["AdditionalProperties"] = {}
            for prop in variable.findall("AdditionalProperties/Property"):
                self.probnet["Variables"][variable_name]["AdditionalProperties"][
                    prop.attrib["name"]
                ] = prop.attrib["value"]
        if variable.find("States/State") is None:
            warnings.warn("States not available for node: " + variable_name)
        else:
            self.probnet["Variables"][variable_name]["States"] = {
                state.attrib["name"]: {
                    prop.attrib["name"]: prop.attrib["value"]
                    for prop in state.findall("AdditionalProperties/Property")
                }
                for state in variable.findall("States/State")
            }

    def add_edge(self, edge):
        """
        Adds Edges to the probnet dict.

        Parameters
        ----------
        edge: <Element Link at Links Node in XML>
            etree Element consisting Variable tag.

        Examples
        -------
        >>> reader = ProbModelXMLReader()
        >>> reader.add_edge(edge)
        """
        var1 = edge.findall("Variable")[0].attrib["name"]
        var2 = edge.findall("Variable")[1].attrib["name"]
        self.probnet["edges"][(var1, var2)] = {}
        self.probnet["edges"][(var1, var2)]["directed"] = edge.attrib["directed"]
        # TODO: check for the case of undirected graphs if we need to add to both elements of the dic for a single edge.
        if edge.find("Comment") is not None:
            self.probnet["edges"][(var1, var2)]["Comment"] = edge.find("Comment").text
        if edge.find("Label") is not None:
            self.probnet["edges"][(var1, var2)]["Label"] = edge.find("Label").text
        if edge.find("AdditionalProperties/Property") is not None:
            self.probnet["edges"][(var1, var2)]["AdditionalProperties"] = {}
            for prop in edge.findall("AdditionalProperties/Property"):
                self.probnet["edges"][(var1, var2)]["AdditionalProperties"][
                    prop.attrib["name"]
                ] = prop.attrib["value"]

    def add_potential(self, potential, potential_dict):
        """
        Adds Potential to the potential dict.

        Parameters
        ----------
        potential: <Element Potential at Potentials node in XML>
            etree Element consisting Potential tag.
        potential_dict: dict{}
            Dictionary to parse Potential tag.

        Examples
        -------
        >>> reader = ProbModelXMLReader()
        >>> reader.add_potential(potential, potential_dict)
        """
        potential_type = potential.attrib["type"]
        potential_dict["type"] = potential_type
        try:
            potential_dict["role"] = potential.attrib["role"]
        except KeyError:
            pass
        if potential.find("Comment") is not None:
            potential_dict["Comment"] = potential.find("Comment").text
        for prop in potential.findall("AdditionalProperties/Property"):
            potential_dict["AdditionalProperties"][prop.attrib["name"]] = prop.attrib[
                "value"
            ]
        if potential_type == "delta":
            potential_dict["Variable"] = potential.find("Variable").attrib["name"]
            if potential.find("State") is not None:
                potential_dict["State"] = potential.find("State").text
            if potential.find("StateIndex") is not None:
                potential_dict["StateIndex"] = potential.find("StateIndex").text
            if potential.find("NumericValue") is not None:
                potential_dict["NumericValue"] = potential.find("NumericValue").text
        else:
            if potential.find("UtilityVariable") is not None:
                potential_dict["UtilityVaribale"] = potential.find(
                    "UtilityVariable"
                ).attrib["name"]
            if len(potential.findall("Variables/Variable")):
                potential_dict["Variables"] = {}
                var_list = []
                for var in potential.findall("Variables/Variable"):
                    var_list.append(var.attrib["name"])
                potential_dict["Variables"][var_list[0]] = var_list[1:]
            if potential.find("Values") is not None:
                potential_dict["Values"] = potential.find("Values").text
            if len(potential.findall("UncertainValues/Value")):
                potential_dict["UncertainValues"] = []
                for value in potential.findall("UncertainValues/Value"):
                    try:
                        potential_dict["UncertainValues"].append(
                            {
                                "distribution": value.attrib["distribution"],
                                "name": value.attrib["name"],
                                "value": value.text,
                            }
                        )
                    except KeyError:
                        potential_dict["UncertainValues"].append(
                            {
                                "distribution": value.attrib["distribution"],
                                "value": value.text,
                            }
                        )
            if potential.find("TopVariable") is not None:
                potential_dict["TopVariable"] = potential.find("TopVariable").attrib[
                    "name"
                ]

            if len(potential.findall("Branches/Branch")):
                potential_dict["Branches"] = []
                for branch in potential.findall("Branches/Branch"):
                    branch_dict = {}
                    if len(branch.findall("States/State")):
                        states = []
                        for state in branch.findall("States/State"):
                            states.append({"name": state.attrib["name"]})
                        branch_dict["States"] = states
                    if branch.find("Potential") is not None:
                        branch_potential = {}
                        self.add_potential(branch.find("Potential"), branch_potential)
                        branch_dict["Potential"] = branch_potential
                    if branch.find("Label") is not None:
                        label = branch.find("Label").text
                        branch_dict["Label"] = label
                    if branch.find("Reference") is not None:
                        reference = branch.find("Reference").text
                        branch_dict["Reference"] = reference
                    if len(branch.findall("Thresholds/Threshold")):
                        thresholds = []
                        for threshold in branch.findall("Thresholds/Threshold"):
                            try:
                                thresholds.append(
                                    {
                                        "value": threshold.attrib["value"],
                                        "belongsTo": threshold.attrib["belongsTo"],
                                    }
                                )
                            except KeyError:
                                thresholds.append({"value": threshold.attrib["value"]})
                        branch_dict["Thresholds"] = thresholds
                    potential_dict["Branches"].append(branch_dict)

            if potential.find("Model") is not None:
                potential_dict["Model"] = potential.find("Model").text
            if len(potential.findall("Subpotentials/Potential")):
                potential_dict["Subpotentials"] = []
                for subpotential in potential.findall("Subpotentials/Potential"):
                    subpotential_dict = {}
                    self.add_potential(subpotential, subpotential_dict)
                    potential_dict["Subpotentials"].append(subpotential_dict)
            if potential.find("Coefficients") is not None:
                potential_dict["Coefficients"] = potential.find("Coefficients").text
            if potential.find("CovarianceMatrix") is not None:
                potential_dict["CovarianceMatrix"] = potential.find(
                    "CovarianceMatrix"
                ).text
            if potential.find("Potential") is not None:
                potential_dict["Potential"] = {}
                self.add_potential(
                    potential.find("Potential"), potential_dict["Potential"]
                )
            if len(potential.findall("NumericVariables/Variable")):
                potential_dict["NumericVariables"] = []
                for variable in potential.findall("NumericVariables/Variable"):
                    potential_dict["NumericVariables"].append(variable.attrib["name"])

    def get_model(self):
        """
        Returns the model instance of the ProbModel.

        Return
        ---------------
        model: an instance of BayesianModel.

        Examples
        -------
        >>> reader = ProbModelXMLReader()
        >>> reader.get_model()
        """
        if self.probnet.get("type") == "BayesianNetwork":
            model = BayesianModel()
            model.add_nodes_from(self.probnet["Variables"].keys())
            model.add_edges_from(self.probnet["edges"].keys())

            tabular_cpds = []
            cpds = self.probnet["Potentials"]
            for cpd in cpds:
                var = list(cpd["Variables"].keys())[0]
                states = self.probnet["Variables"][var]["States"]
                evidence = cpd["Variables"][var]
                evidence_card = [
                    len(self.probnet["Variables"][evidence_var]["States"])
                    for evidence_var in evidence
                ]
                arr = list(map(float, cpd["Values"].split()))
                values = np.array(arr)
                values = values.reshape((len(states), values.size // len(states)))
                tabular_cpds.append(
                    TabularCPD(var, len(states), values, evidence, evidence_card)
                )

            model.add_cpds(*tabular_cpds)

            variables = model.nodes()
            for var in variables:
                for prop_name, prop_value in self.probnet["Variables"][var].items():
                    model.nodes[var][prop_name] = prop_value
            edges = model.edges()

            if nx.__version__.startswith("1"):
                for edge in edges:
                    for prop_name, prop_value in self.probnet["edges"][edge].items():
                        model.edge[edge[0]][edge[1]][prop_name] = prop_value
            else:
                for edge in edges:
                    for prop_name, prop_value in self.probnet["edges"][edge].items():
                        model.adj[edge[0]][edge[1]][prop_name] = prop_value
            return model
        else:
            raise ValueError("Please specify only Bayesian Network.")
